#include <iostream>
#include <cstdlib>
#include <fstream>
#include <cmath>
#include <ctime>
#include "image2d.h"
using namespace std;

//default constructor.
Image2D::Image2D() {
	NX = 0;
	NY = 0;
	Q = 0;

	pixelValue = NULL;
}

//constructor.
Image2D::Image2D(int tmpNX, int tmpNY, int tmpQ) {
	int i, j;

	NX = tmpNX;
	NY = tmpNY;
	Q = tmpQ;

	pixelValue = new double* [NX];
	for(i=0; i<NX; i++) {
		pixelValue[i] = new double[NY];
		for(j=0; j<NY; j++)
			pixelValue[i][j] = 0;
	}
}

Image2D::Image2D(int tmpNX, int tmpNY, int tmpQ, double grayVal) {
	int i, j;

	NX = tmpNX;
	NY = tmpNY;
	Q = tmpQ;

	pixelValue = new double* [NX];
	for(i=0; i<NX; i++) {
		pixelValue[i] = new double[NY];
		for(j=0; j<NY; j++)
			pixelValue[i][j] = grayVal;
	}
}


//copy constructor.
Image2D::Image2D(const Image2D& oldImage) {
	int i, j;

	NX = oldImage.NX;
	NY = oldImage.NY;
	Q = oldImage.Q;

	pixelValue = new double* [NX];
	for(i=0; i<NX; i++) {
		pixelValue[i] = new double[NY];
		for(j=0; j<NY; j++)
			pixelValue[i][j] = oldImage.pixelValue[i][j];
	}
}

//overload operator=
Image2D& Image2D::operator=(const Image2D& oldImage) {
  if(this != &oldImage) {
    int i, j;
    if(NX == oldImage.NX && NY == oldImage.NY && pixelValue != NULL) {
      for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++)
	  pixelValue[i][j] = oldImage.pixelValue[i][j];
      }
    }
    else {
      if(pixelValue != NULL) {
	for(i=0; i<NX; i++)
	  delete [] pixelValue[i];
	delete [] pixelValue;
      }
      NX = oldImage.NX;
      NY = oldImage.NY;
      Q = oldImage.Q;
      
      pixelValue = new double* [NX];
      for(i=0; i<NX; i++) {
	pixelValue[i] = new double[NY];
	for(j=0; j<NY; j++)
	  pixelValue[i][j] = oldImage.pixelValue[i][j];
      }
    }
  }
  return *this;
}
/*
Image2D Image2D::expand(int c, int r, double val)const {
    int nx,ny,q;
    int i,j;
    nx = 2*c + NX; ny = 2*r + NY;
    q = Q;
    Image2D new_im(nx,ny,q);
    for(i=0; i<nx; i++) {
	for(j=0; j<ny; j++) {
	    if(i < c || i >= NX + c || j < r || j >= NY + r)
		new_im(i,j) = val;
	    else
		new_im(i,j) = pixelValue[i-c][j-r];
	}
    }
    return new_im;
}
*/
Image2D::~Image2D() {
  if(pixelValue != NULL) {
    int i;
    
    for(i=0; i<NX; i++)
      delete [] pixelValue[i];
    delete [] pixelValue;
  }
}

void Image2D::getImageInfo(int& rows, int& cols, int& levels) const {
	rows = NX;
	cols = NY;
	levels = Q;
}

bool Image2D::inBounds(int rows, int cols) const {
	return ((rows >= 0) && (rows < NX) && (cols >= 0) && (cols < NY));
}

Image2D Image2D::getSubImage(int upperLeftR, int upperLeftC, int
			     lowerRightR, int lowerRightC) const {
  if(inBounds(upperLeftR, upperLeftC) && inBounds(lowerRightR, lowerRightC)
     && (upperLeftR < lowerRightR) && (upperLeftC < lowerRightC)) {
    
    int i,j;
    Image2D temp(lowerRightR - upperLeftR,lowerRightC - upperLeftC, Q);
		for(i=0; i<temp.NX; i++)
		  for(j=0; j<temp.NY; j++)
		    temp.pixelValue[i][j] = 
		      pixelValue[upperLeftR + i][upperLeftC + j];
		return temp;
  }
  else {
    cout << "error in getSubImage\n";
    exit(1);
  }
}

double Image2D::mean() const {
	double total = 0;
	int i, j;
	for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++)
			total += pixelValue[i][j];
	}
	double dx = 1.0/(NX-1), dy = 1.0/(NY-1);
	return total*dx*dy;
}

double Image2D::min() const {
    double m = Q+1;
    int i, j;
    for(i=0; i<NX; i++) {
        for(j=0; j<NY; j++) {
           if(pixelValue[i][j] < m)
               m = pixelValue[i][j];
        }
    }
    return m;
}
double Image2D::max() const {
    double m = -1;
    int i, j;
    for(i=0; i<NX; i++) {
        for(j=0; j<NY; j++) {
           if(pixelValue[i][j] > m)
               m = pixelValue[i][j];
        }
    }
    return m;
}

double& Image2D::operator()(int i, int j) {
	if(inBounds(i,j))
		return pixelValue[i][j];
	else {
		cout << "error in operator()\n";
		exit(1);
	}
}

const double& Image2D::operator()(int i, int j) const {
	if(inBounds(i,j))
		return pixelValue[i][j];
	else {
		cout << " error in operator()\n";
		exit(1);
	}
}
void Image2D::reflect(int n) {
    if(n>0 && n < NX/3 && n < NY/3) {
	int i,j;
	//reflect the vertical border.
	for(i=n;i<NX-n;i++) {
            for(j=0;j<n;j++) {
		pixelValue[i][j] = pixelValue[i][n];
		pixelValue[i][NY-1-j] = pixelValue[i][NY-1-n];
            }
	}
	//reflect the horizontal border.
	for(j=0;j<NY;j++) {
	    for(i=0;i<n;i++) {
		pixelValue[i][j] = pixelValue[n][j];
		pixelValue[NX-1-i][j] = pixelValue[NX-1-n][j];
	    }
	}
    }
    else {
	cout << "Error in void reflect(int).\n";
        exit(1);
    }
}

Image2D Image2D::expand(int n) const {
  if(n > 0) {
    int nx, ny, i, j;
    nx = NX+2*n;
    ny = NY+2*n;
    Image2D temp(nx,ny,Q);
    for(i=0; i<nx; i++) {
      for(j=0; j<ny; j++) {
	if(i < n || i >= nx - n || j < n || j >= ny - n) {
	  temp(i,j) = 0;
	}
	else
	  temp(i,j) = pixelValue[i-n][j-n];
      }
    }
    return temp;
  }
  else
    return *this;
}

Image2D& Image2D::restrict() {
    int i,j;
    for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++) {
	    if(pixelValue[i][j] < 0)
		pixelValue[i][j] = 0;
	    else if(pixelValue[i][j] > Q)
		pixelValue[i][j] = Q;
	}
    }
    return *this;
}

Image2D& Image2D::operator -=(const Image2D& rhs) {
	if(NX != rhs.NX || NY != rhs.NY) {
	  cout << "error in operator-=\n";
	  exit(1);
	}
	int i, j;
	for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++) {
			pixelValue[i][j] -= rhs.pixelValue[i][j];
		}
	}
	return *this;
}

Image2D& Image2D::operator -=(double val) {
	int i, j;
	for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++) {
			pixelValue[i][j] -= val;
		}
	}
	return *this;
}


Image2D& Image2D::operator +=(const Image2D& rhs) {
	if(NX != rhs.NX || NY != rhs.NY) {
	  cout << "error in operator+=\n";
	  exit(1);
	}
	int i, j;
	for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++) {
			pixelValue[i][j] += rhs.pixelValue[i][j];
		}
	}
	return *this;
}

Image2D& Image2D::operator +=(double val) {
	int i, j;
	for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++) {
			pixelValue[i][j] += val;
		}
	}
	return *this;
}

Image2D& Image2D::operator *=(const Image2D& rhs) {
	if(NX != rhs.NX || NY != rhs.NY) {
	  cout << "error in operator*=\n";
	  exit(1);
	}
	int i, j;
	for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++) {
			pixelValue[i][j] *= rhs.pixelValue[i][j];
		}
	}
	return *this;
}

Image2D& Image2D::operator *=(double val) {
	int i, j;
	for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++) {
			pixelValue[i][j] *= val;
		}
	}
	return *this;
}


Image2D& Image2D::operator /=(const Image2D& rhs) {
  if(NX != rhs.NX || NY != rhs.NY) {
    cout << "error in operator/=\n";
    exit(1);
  }
  int i, j;
  for(i=0; i<NX; i++) {
    for(j=0; j<NY; j++) {
      if(rhs.pixelValue[i][j] == 0)
	pixelValue[i][j] /= rhs.pixelValue[i][j] + 1e-6;
      else
	pixelValue[i][j] /= rhs.pixelValue[i][j];
      
    }
  }
  return *this;
}

Image2D& Image2D::operator /=(double val) {
  if(val > 0) {
    int i, j;
    for(i=0; i<NX; i++) {
      for(j=0; j<NY; j++) {
	pixelValue[i][j] /= val;
      }
    }
    return *this;
  }
  else {
    cout << "error in operator/=\n";
    exit(1);
  }
}

//---------NON MEMBER FUNCTIONS--------------
Image2D operator-(const Image2D& rhs) {
    Image2D temp(rhs);
    return -1*temp;
}

Image2D operator-(const Image2D& lhs, const Image2D& rhs) {
	Image2D temp(lhs);
	temp -= rhs;
	return temp;
}

Image2D operator-(const Image2D& lhs, double val) {
	Image2D temp(lhs);
	temp -= val;
	return temp;
}

Image2D operator-(double val, const Image2D& rhs) {
	Image2D temp(-rhs);
	temp += val;
	return temp;
}

Image2D operator+(const Image2D& lhs, const Image2D& rhs) {
	Image2D temp(lhs);
	temp += rhs;
	return temp;
}

Image2D operator+(const Image2D& lhs, double val) {
	Image2D temp(lhs);
	temp += val;
	return temp;
}

Image2D operator+(double val, const Image2D& rhs) {
	return rhs+val;
}

Image2D operator*(const Image2D& lhs, const Image2D& rhs) {
	Image2D temp(lhs);
	temp *= rhs;
	return temp;
}

Image2D operator*(const Image2D& lhs, double val) {
	Image2D temp(lhs);
	temp *= val;
	return temp;
}

Image2D operator*(double val, const Image2D& rhs) {
	Image2D temp(rhs);
	temp *= val;
	return temp;
}

Image2D operator/(const Image2D& rhs, const Image2D& lhs) {
	Image2D temp(rhs);
	temp /= lhs;
	return temp;
}

Image2D operator/(const Image2D& top, double val) {
	Image2D temp(top);
	temp /= val;
	return temp;
}

Image2D operator/(double val, const Image2D& bottom) {
    int nx,ny,q,i,j;
    double eps = 1e-6;
    bottom.getImageInfo(nx,ny,q);
    Image2D temp(nx,ny,q,val);
    for(i=0; i<nx; i++) {
	for(j=0; j<ny; j++) {
	    if(bottom(i,j) == 0)
		temp(i,j) /= bottom(i,j) + eps;
	    else
		temp(i,j) /= bottom(i,j);
	}
    }
    return temp;
}


Image2D abs(const Image2D& im) {
    Image2D temp(im);
    int nx,ny,q,i,j;
    temp.getImageInfo(nx,ny,q);
    for(i=0; i<nx; i++) {
	for(j=0; j<ny; j++) {
	    temp(i,j) = fabs(temp(i,j));
	}
    }
    return temp;
}

Image2D heavisideEps(const Image2D& phi, double eps) {
    int nx,ny,q,i,j;
    Image2D H(phi);
    H.getImageInfo(nx,ny,q);
    for(i=0; i<nx; i++) {
	for(j=0; j<ny; j++) {
	    H(i,j) = 0.5*(1 + (2.0/PI)*atan(phi(i,j)/eps));
	}
    }
    return H;
}

Image2D deltaEps(const Image2D& phi, double eps) {
    int nx,ny,q,i,j;
    Image2D D(phi);
    D.getImageInfo(nx,ny,q);
    for(i=0; i<nx; i++) {
	for(j=0; j<ny; j++) {
	    D(i,j) = (1/PI)*eps/(eps*eps + phi(i,j)*phi(i,j));
	}
    }
    return D;
}

Image2D pow(const Image2D& f, double exp) {
    int nx,ny,q,i,j;
    Image2D f_to_exp(f);
    f_to_exp.getImageInfo(nx,ny,q);
    for(i=0; i<nx; i++) {
	for(j=0; j<ny; j++) {
	    f_to_exp(i,j) = pow(f(i,j),exp);
	}
    }
    return f_to_exp;
}

Image2D sqrt(const Image2D& f) {
    int nx,ny,q,i,j;
    Image2D sqrt_f(f);
    sqrt_f.getImageInfo(nx,ny,q);
    for(i=0; i<nx; i++) {
	for(j=0; j<ny; j++) {
	    sqrt_f(i,j) = sqrt(f(i,j));
	}
    }
    return sqrt_f;
}

double norm(const Image2D& f, const string& type) {
    int i,j,nx,ny,q;
    double result = 0;
    f.getImageInfo(nx,ny,q);
    //if(type == "inf") {
	for(i=0; i<nx; i++) {
	    for(j=0; j<ny; j++) {
		if(fabs(f(i,j)) > result)
		    result = fabs(f(i,j));
	    }
	}
	//}
    return result;
}

void readImage2D(char fname[], Image2D& image, bool type) {
	int i, j;
	int NX, NY, Q;
	
	char header [100], *ptr;
	ifstream ifp;

	ifp.open(fname, ios::in);

	if (!ifp) {
		cout << "Can't read image: " << fname << endl;
		exit(1);
	}

	// read header

	ifp.getline(header,100,'\n');
	if ((header[0]!=80) ||    /* 'P' */
		((header[1]!=53) && (header[1] != 50))) {   /* '5' || '2'*/
		cout << "Image " << fname << " is not PGNY" << endl;
		exit(1);
	}

	ifp.getline(header,100,'\n');
	while(header[0]=='#')
		ifp.getline(header,100,'\n');

	NY=strtol(header,&ptr,0);
	NX=atoi(ptr);

	ifp.getline(header,100,'\n');
	Q=strtol(header,&ptr,0);
	if(type) {
		char *charImage;
		charImage = new char [NY*NX];

		ifp.read(charImage, (NY*NX)*sizeof(unsigned char));

		if (ifp.fail()) {
			cout << "Image " << fname << " has wrong size" << endl;
			exit(1);	
		}
		ifp.close();

		//
		// Convert the unsigned characters to integers
		//

		double val;

		for(i=0; i<NX; i++)
			for(j=0; j<NY; j++) {
				val = (double)((unsigned char)charImage[i*NY+j]);
				image(i,j) = val;     
			}

		delete [] charImage;
	
	}
	else {
		for(i=0; i<NX; i++) {
			for(j=0; j<NY; j++) {
				if(!ifp.eof()) {
					ifp >> image(i,j);
				}
			}
		}
	}
	
}

void writeImage2D(const char fname[], const Image2D& image, bool type) {
	int i, j;
	int NX, NY, Q;
	
	ofstream ofp;

	image.getImageInfo(NX, NY, Q);

	// convert the integer values to unsigned char

	ofp.open(fname, ios::out);

	if (!ofp) {
		cout << "Can't open file: " << fname << endl;
		exit(1);
	}

	if(type) {
		ofp << "P5" << endl;
		ofp << NY << " " << NX << endl;
		ofp << Q << endl;
		char *charImage;
		charImage = new char [NY*NX];
		double val;

		for(i=0; i<NX; i++)
			for(j=0; j<NY; j++) {
				val = image(i,j);
				charImage[i*NY+j]=(char)val;
			}
		ofp.write(charImage, (NY*NX)*sizeof(unsigned char));

		if (ofp.fail()) {
			cout << "Can't write image " << fname << endl;
			exit(0);
		}

		ofp.close();

		delete [] charImage;
	}
	else {
		ofp << "P2" << endl;
		ofp << NY << " " << NX << endl;
		ofp << Q << endl;
		for(i=0;i<NX;i++) {
			for(j=0;j<NY;j++) {
				ofp << (unsigned int)image(i,j) << " ";
			}
			ofp << endl;
		}
	}
}

void readImage2DHeader(char fname[], int& NX, int& NY, int& Q, bool& type) {

	char header [100], *ptr;
	ifstream ifp;

	ifp.open(fname, ios::in);

	if (!ifp) {
		cout << "Can't read image: " << fname << endl;
		exit(1);
	}

	// read header

	type = false; // PGM

	ifp.getline(header,100,'\n');
	if ((header[0] == 80) &&  /* 'P' */
		(header[1]== 53)) {  /* '5' */
		type = true;
	}
	else if ( (header[0] == 80) &&  /* 'P' */
		(header[1] == 50) ) {        /* '2' */
		type = false;
	} 
	else {
		cout << "Image " << fname << " is not PGM or PPM" << endl;
		exit(1);
	}

	ifp.getline(header,100,'\n');
	while(header[0]=='#')
		ifp.getline(header,100,'\n');

	NY=strtol(header,&ptr,0);
	NX=atoi(ptr);

	ifp.getline(header,100,'\n');

	Q=strtol(header,&ptr,0);

	ifp.close();
}

