#ifndef IMAGE2D_H
#define IMAGE2D_H

#include <cstring>
//using namespace std;

const double PI = 3.14159265;
class Image2D {
 public:
   Image2D();
   Image2D(int, int, int);
   Image2D(int,int,int,double);
   Image2D(const Image2D&);
   ~Image2D();
   //member functions.
   void getImageInfo(int&, int&, int&) const;
   bool inBounds(int, int) const;
   Image2D getSubImage(int,int,int,int) const;
   double mean() const;
   double min() const;
   double max() const;
   void reflect(int n);
   Image2D expand(int n)const;
   Image2D& restrict();
   //overloaded operators.
   double& operator()(int,int);
   const double& operator()(int,int) const;
   Image2D& operator=(const Image2D&);
   Image2D& operator-=(const Image2D&);
   Image2D& operator-=(double);
   Image2D& operator+=(const Image2D&);
   Image2D& operator+=(double);
   Image2D& operator*=(const Image2D&);
   Image2D& operator*=(double);
   Image2D& operator/=(const Image2D&);
   Image2D& operator/=(double);

 private:
   int NX, NY, Q;
   double **pixelValue;
};

//non-member functions.
void readImage2D(char fileName[], Image2D&, bool type);
void writeImage2D(const char fileName[], const Image2D&, bool type);
void readImage2DHeader(char fname[], int& NX, int& NY, int& Q, bool& type);
Image2D operator-(const Image2D&);
Image2D operator-(const Image2D&, const Image2D&); 
Image2D operator-(const Image2D&, double);
Image2D operator-(double, const Image2D&);
Image2D operator+(const Image2D&, const Image2D&);
Image2D operator+(const Image2D&, double);
Image2D operator+(double, const Image2D&);
Image2D operator*(const Image2D&, const Image2D&);
Image2D operator*(const Image2D&, double);
Image2D operator*(double, const Image2D&);
Image2D operator/(const Image2D&, const Image2D&);
Image2D operator/(const Image2D&, double);
Image2D operator/(double, const Image2D&);
Image2D abs(const Image2D&);
Image2D heavisideEps(const Image2D& phi, double eps);
Image2D deltaEps(const Image2D& phi, double eps);
Image2D pow(const Image2D& f, double exp);
Image2D sqrt(const Image2D& f);
//double norm(const Image2D& f, const string& type);

#endif
