#include <iostream>
#include <cstdlib>
#include <fstream>
#include <cmath>
#include <ctime>
#include "image3d.h"
#include "image2d.h"

using namespace std;

//default constructor.
Image3D::Image3D() {
    NX = 0;
    NY = 0;
    NZ = 0;
    Q = 0;
    images = NULL;
}

//constructor.
Image3D::Image3D(int nx, int ny, int nz, int q) {
    int i, j, k;

    NX = nx;
    NY = ny;
    NZ = nz;
    Q = q;
	
    images = new Image2D [NZ];
    for(k=0; k<NZ; k++)
	images[k] = Image2D(NX,NY,Q);
}

Image3D::Image3D(int nx, int ny, int nz, int q, double grayVal) {
    int i, j, k;

    NX = nx;
    NY = ny;
    NZ = nz;
    Q = q;
	
    images = new Image2D [NZ];
    for(k=0; k<NZ; k++)
	images[k] = Image2D(NX,NY,Q,grayVal);
}


//copy constructor.
Image3D::Image3D(const Image3D& oldImage) {
    int i, j, k;
    NX = oldImage.NX;
    NY = oldImage.NY;
    NZ = oldImage.NZ;
    Q = oldImage.Q;

    images = new Image2D [NZ];
    for(k=0; k<NZ; k++) {
		images[k] = oldImage.images[k];
    }
}

//overload operator=
Image3D& Image3D::operator=(const Image3D& oldImage) {
    if(this != &oldImage) {
	int i, j, k;
	if(NX != oldImage.NX || NY != oldImage.NY || NZ != oldImage.NZ) {
	    
	    if(images != NULL)
		delete [] images;
	    NX = oldImage.NX;
	    NY = oldImage.NY;
	    NZ = oldImage.NZ;
	    Q = oldImage.Q;
	    images = new Image2D [NZ];
	    for(k=0; k<NZ; k++) {
			images[k] = oldImage.images[k];
	    }
	}
	else {
	    for(k=0; k<NZ; k++) {
			images[k] = oldImage.images[k];
	    }
	}
    }
    return *this;
}


Image3D::~Image3D() {
    if(images != NULL) {
	delete [] images;
    }
}

void Image3D::getImageInfo(int& nx, int& ny, int& nz, int& q) const {
    nx = NX;
    ny = NY;
    nz = NZ;
    q = Q;
} 

bool Image3D::inBounds(int nx, int ny, int nz) const {
	return ((nx >= 0) && (nx < NX) && (ny >= 0) 
		&& (ny < NY) && (nz >= 0) && (nz < NZ));
}

double Image3D::mean() const {
    double total = 0;
    int i, j, k;
    for(k=0; k<NZ; k++) {
		total += images[k].mean();
    }
    if(NZ > 0)
	return total/NZ;
}

double Image3D::min() const {
    double m = Q,mi;
    int i, j, k;
    for(k=0; k<NZ; k++) {
	mi = images[k].min();
	if(mi < m)
	    m = mi;
    }
    return m;
}
double Image3D::max() const {
    double m = -1,ma;
    int i, j, k;
    for(k=0; k<NZ; k++) {
	ma = images[k].max();
	if(ma > m)
	    m = ma;
    }
    return m;
}

double& Image3D::operator()(int i, int j, int k) {
	if(inBounds(i,j,k))
		return images[k](i,j);
	else {
		cout << "out of bounds\n";
		exit(1);
	}
}

double Image3D::operator()(int i, int j, int k) const {
	if(inBounds(i,j,k))
		return images[k](i,j);
	else {
		cout << " out of bounds\n";
		exit(1);
	}
}
/*
Image2D& Image3D::slicing(int k) {
    if(k>=0 && k<NZ)
	return images[k];
    else {
	cout << "error in Slicing\n";
	exit(1);
    }
}
*/
const Image2D& Image3D::slicing(int k) const{
    if(k>=0 && k<NZ)
	return images[k];
    else {
	cout << "error in Slicing\n";
	exit(1);
    }
}

void Image3D::reflect(int n) {
    int i,j,k;
    // i is in front/back direction.
    // j is in the left/right direction.
    // k is in the top/bottom direction.
    for(k=0; k<NZ; k++)
	images[k].reflect(n);
    for(k=0; k<n; k++) {
	images[k] = images[n];
	images[NZ-1-k] = images[NZ-1-n];
    }
}
Image3D& Image3D::operator -=(const Image3D& lhs) {
    if(NX != lhs.NX || NY != lhs.NY || NZ != lhs.NZ) {
	//throw exception.
	exit(1);
    }
    int i, j, k;
    for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++) {
	    for(k=0; k<NZ; k++) {
		images[k](i,j) -= lhs(i,j,k);
	    }
	}
    }
    return *this;
}

Image3D& Image3D::operator -=(double lhs) {
    int i, j, k;
    for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++) {
	    for(k=0; k<NZ; k++) {
		images[k](i,j) -= lhs;
	    }
	}
    }
    return *this;
}


Image3D& Image3D::operator +=(const Image3D& lhs) {
    if(NX != lhs.NX || NY != lhs.NY || NZ != lhs.NZ) {
	//throw exception.
	exit(1);
    }
    int i, j, k;
    for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++) {
	    for(k=0; k<NZ; k++) {
		images[k](i,j) += lhs(i,j,k);
	    }
	}
    }
    return *this;
}

Image3D& Image3D::operator +=(double lhs) {
    int i, j, k;
    for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++) {
	    for(k=0; k<NZ; k++) {
		images[k](i,j) += lhs;
	    }
	}
    }
    return *this;
}

Image3D& Image3D::operator *=(const Image3D& lhs) {
    if(NX != lhs.NX || NY != lhs.NY || NZ != lhs.NZ) {
	//throw exception.
	exit(1);
    }
    int i, j, k;
    for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++) {
	    for(k=0; k<NZ; k++) {
		images[k](i,j) *= lhs.images[k](i,j);
	    }
	}
    }
    return *this;
}

Image3D& Image3D::operator *=(double val) {
    int i, j, k;
    for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++) {
	    for(k=0; k<NZ; k++) {
		images[k](i,j) *= val;
	    }
	}
    }
    return *this;
}


Image3D& Image3D::operator /=(const Image3D& lhs) {
    if(NX != lhs.NX || NY != lhs.NY || NZ != lhs.NZ) {
	//throw exception.
	exit(1);
    }
    int i, j, k;
    for(i=0; i<NX; i++) {
	for(j=0; j<NY; j++) {
	    for(k=0; k<NZ; k++) {
		images[k](i,j) /= lhs.images[k](i,j) + 10e-5;
	    }
	}
    }
    return *this;
}


//---------NON MEMBER FUNCTIONS--------------


Image3D operator-(const Image3D& rhs, const Image3D& lhs) {
	Image3D temp(rhs);
	temp -= lhs;
	return temp;
}

Image3D operator-(const Image3D& rhs, double val) {
	Image3D temp(rhs);
	temp -= val;
	return temp;
}

Image3D operator-(double val, const Image3D& lhs) {
	return lhs-val;
}

Image3D operator+(const Image3D& rhs, const Image3D& lhs) {
	Image3D temp(rhs);
	temp += lhs;
	return temp;
}

Image3D operator+(const Image3D& rhs, double val) {
	Image3D temp(rhs);
	temp += val;
	return temp;
}

Image3D operator+(double val, const Image3D& lhs) {
	return lhs+val;
}

Image3D operator*(const Image3D& rhs, const Image3D& lhs) {
	Image3D temp(rhs);
	temp *= lhs;
	return temp;
}

Image3D operator*(const Image3D& rhs, double val) {
	Image3D temp(rhs);
	temp *= val;
	return temp;
}

Image3D operator*(double val, const Image3D& lhs) {
	return lhs*val;
}

Image3D operator/(const Image3D& rhs, const Image3D& lhs) {
	Image3D temp(rhs);
	temp /= lhs;
	return temp;
}

void readImage3DHeader(char listFile[], int& NX, int& NY, int& NZ, int& Q, bool& type) {
    int i;
    char fname [500], *ptr, header[500];
    ifstream ifp;
    ifstream inFile;
    int counter = 0;
    ifp.open(listFile);
    
	
    if (!ifp) {
	cout << "Can't read list File: " << listFile << endl;
	exit(1);
    }
	
    //cout << "Can't read list File: " << listFile << endl;
    //counting the number of slices from listFile
    ifp.getline(fname,500,'\n');
    counter = 1;
    while(!ifp.eof()) {
	ifp.getline(header,500,'\n');
	counter++;
    }
	
//    NZ = counter;
//    ifp.close();
//  inFile.open(fname, ios::in);  //this doesn;t work so I commented out
//  inFile.open(fname); 
//	ifp.open(fname, ios::in);
//  if (!ifp) {
//	cout << "Can't read image file: " << fname << endl;
//	exit(1);
//  }
//	cout<<" I'm in readImage3DHeader "<<endl;  ok 
    //cout << NX << endl;
    type = false; // PGM
    inFile.getline(header,500,'\n');
	cout<<header[1]<<endl;
    if ((header[0] == 80) &&  /* 'P' */
	(header[1]== 53)) {  /* '5' */
	type = true;
    }
    else if ( (header[0] == 80) &&  /* 'P' */
	      (header[1] == 50) ) {        /* '2' */
	type = false;
    } 
    else {
	cout << "Image " << fname << " is not PGM or PPM" << endl;
	exit(1);
    }
    
    inFile.getline(header,500,'\n');
    while(header[0]=='#')
	inFile.getline(header,500,'\n');
    
    NY=strtol(header,&ptr,0);
    NX=atoi(ptr);
    
    inFile.getline(header,500,'\n');
    
    Q=strtol(header,&ptr,0);
    
    inFile.close();

	cout<<" End of readHeader "<<endl;
}

void readImage3D(char listName[], Image3D& image, bool type) {
    
	int i, j, k;
    int NX, NY, NZ, Q;
    
    char header [100], **fileList;
    ifstream ifp;
    ifstream *inFile;
    
    image.getImageInfo(NX,NY,NZ,Q);
    fileList = new char*[NZ];
    inFile = new ifstream[NZ];
    for(k=0; k<NZ; k++) {
	fileList[k] = new char[100];
    }
    ifp.open(listName, ios::in);
    
	
    if (!ifp) {
	cout << "Can't read List File: " << listName << endl;
	exit(1);
    }
    k = 0;
    while(!ifp.eof()) {
	ifp.getline(fileList[k],500,'\n');
	k++;
	cout<<k<<" ";
    }
//	cout<<" Test 1 " <<endl;
    ifp.close();
    
    for(k=0; k<NZ; k++) {
	inFile[k].open(fileList[k], ios::in);
	if (!inFile[k]) {
	    cout << "Can't read image file: " << fileList[k] << endl;
	    exit(1);
	}
	inFile[k].getline(header,100,'\n');
	if ((header[0]!=80) ||    /* 'P' */
	    ((header[1]!=53) && (header[1] != 50))) {   /* '5' */
	    cout << "Image " << fileList[k] << " is not PGM" << endl;
	    exit(1);
	}
	
	inFile[k].getline(header,100,'\n');
	while(header[0]=='#')
	    inFile[k].getline(header,100,'\n');
	
	//NX = strtol(header, &ptr, 0);
	//NY=atoi(ptr);
	
	inFile[k].getline(header,100,'\n');
	//Q=strtol(header,&ptr,0);
	if(type) {
	    char *charImage;
	    charImage = new char [NX*NY];
	    
	    inFile[k].read(charImage, (NX*NY)*sizeof(unsigned char));
	    
	    if (inFile[k].fail()) {
		cout << "Image " << fileList[k] << " has wrong size" << endl;
		exit(1);	
	    }
	    
	    
	    //
	    // Convert the unsigned characters to integers
	    //
	    
	    double val;
	    for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++) {
		    val = (double)charImage[i*NY+j];
		    image(i,j,k) = val;     
		}
	    }
	    delete [] charImage;	
	}

	else {
	    
	    for(i=0; i<NX; i++) {
		for(j=0; j<NY; j++) {		
		    if(!inFile[k].eof()) {
			inFile[k] >> image(i,j,k);
		    }
		}
	    }
	}
	inFile[k].close();
    }
    for(k=0; k<NZ; k++)
	delete [] fileList[k];
    delete [] fileList;
    delete [] inFile;
	
	cout<<" I'm in readImage  3D"<<endl;
}


/*
void readImage3D_PGM2(char listName[], Image3D& image, bool type){
	 
	int i, j, k;
//    int NX, NY, NZ, Q;
	int NX = multilayer3D.dimensions[0], NY = multilayer3D.dimensions[1], NZ = multilayer3D.dimensions[2];
    cout<<NX<<NY<<NZ<<endl;
    
    char header [1000], **fileList;
	ifstream ifp;
    ifstream *inFile;
    
    image.getImageInfo(NX,NY,NZ,Q);
	fileList = new char*[NZ];
    inFile = new ifstream[NZ];
    for(k=0; k<NZ; k++) {
	fileList[k] = new char[1000];
    }
    ifp.open(listName, ios::in);
    
	
    if (!ifp) {
	cout << "Can't read List File: " << listName << endl;
	exit(1);
    }
    k = 0;
    while(!ifp.eof()) {
	ifp.getline(fileList[k],1000,'\n');
	k++;
    }
	cout<<" Hi " <<endl;
    ifp.close();
    
    for(k=0; k<NZ; k++) {
		
		inFile[k].open(fileList[k], ios::in);
		
		if (!inFile[k]) {
		  cout << "Can't read image file: " << fileList[k] << endl;
		  exit(1);
		}
		inFile[k].getline(header,1000,'\n'); // P2
		inFile[k].getline(header,1000,'\n'); // dimx  dimy
		inFile[k].getline(header,1000,'\n'); // max gray value e.g. 255

		if(type) {
			char *charImage;
			charImage = new char [NX*NY];
	    
			inFile[k].read(charImage, (NX*NY)*sizeof(unsigned char));
	    
			if (inFile[k].fail()) {
				cout << "Image " << fileList[k] << " has wrong size" << endl;
				exit(1);	
			}
	    
	    
	    //
	    // Convert the unsigned characters to integers
	    //
	    
		    double val;
			for(i=0; i<NX; i++) {
				for(j=0; j<NY; j++) {
					val = (double)charImage[i*NY+j];
					image(i,j,k) = val;     
				}
			}
			delete [] charImage;	
		}

		else {
	    
			for(i=0; i<NX; i++) {
				for(j=0; j<NY; j++) {		
					if(!inFile[k].eof()) {
					inFile[k] >> image(i,j,k);
					}
				}
			}
		}
		inFile[k].close();
	} // end of for loop

    for(k=0; k<NZ; k++)
	delete [] fileList[k];
    delete [] fileList;
    delete [] inFile;
	cout<<" inside read3D_PGM2 "<<endl;
	
}
*/

/*
void writeImage3D(Image2D& image){
    
	int i, j, k;
	int NX = dimensions[0], NY = dimensions[1], NZ = dimensions[2];
	//int NX, NY, NZ, Q;
	
	//image.getImageInfo(NX,NY,NZ,Q);

	for(k=0; k<NZ; k++) {
		char iternum[30];
	    sprintf(iternum, "%d", k+1);
		char filename[30] = "segmavg";
		strcat(filename,iternum);
		strcat(filename,".dat");
		ofstream fout(filename);
		int counter=1;
		for(i=0;i<NX;i++){
			for(j=0;j<NY;j++){
				fout<<(int)image(i,j)<<" ";
				if(counter%17==0)
					fout<<endl;
				counter++;
			}
		}
		fout.close();

	}

}


void Neumann(Image3D& phi){
	int i,j,k;
	int NX, NY, NZ, Q;
	
	image.getImageInfo(NX,NY,NZ,Q);
	for(i=1;i<NX+1;i++){
		for(k=1;k<NZ+1;k++){
				phi(i,0,k) = phi(i,1,k);
				phi(i,NY+1,k) = phi(i,NY,k);
		}
	}

	for(j=1;j<NY+1;j++){
		for(k=1;k<NZ+1;k++){
			phi(0,j,k) = phi(1,j,k);
			phi(NX+1,j,k) = phi(NX,j,k);
		}
	}
			
	for(i=1;i<NX+1;i++){
		for(j=1;j<NY+1;j++){
			phi(i,j,0) = phi(i,j,1);
			phi(i,j,NZ+1) = phi(i,j,NZ);
		}
	}

	phi(0,0,0) = phi(1,1,1);
	phi(0,0,NZ+1) = phi(1,1,NZ);
	phi(0,NY+1,0) = phi(1,NY,1);
	phi(NX+1,0,0) = phi(NX,1,1);

	phi(0,NY+1,NZ+1) = phi(1,NY,NZ);
	phi(NX+1,0,NZ+1) = phi(NX,1,NZ);
	phi(NX+1,NY+1,0) = phi(NX,NY,1);
	phi(NX+1,NY+1,NZ+1) = phi(NX,NY,NZ);

}

*/